# Job abstraction to manage simultaneous installs
#
# Copyright 2006-2007 Novell, Inc.
# Author: Charles Coffing <ccoffing@novell.com>
#
# This software may be freely redistributed under the terms of the GNU
# general public license.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import logging
import os
import signal
import sys
import tempfile

from exceptions import *
from FullVirtGuest import FullVirtGuest
from log import log, init_log
from ParaVirtGuest import ParaVirtGuest
import caps
import console
import msg
import paths
import util
import xend

class Job(object):
    """
    Caller-visible attributes:
      Job.XYZ -- state names
      job.jobid -- The unique job ID (if backgrounded)
      job.state -- The current state in the installation process
      job.guest -- The VM instance (see XenGuest.py)
      job.guest.options -- The VM settings (see options.py)
    """

    (SETTINGS,
     PREP_INSTALL,
     INSTALL,
     POST_INSTALL,
     RUNTIME_CONF,
     RUNTIME,
     FINISHED,
     FAILED,
     CANCELED) = range(9)

    def __init__(self, background, debug=False):
        """Creates a job."""
        self._states = (
            (Job.SETTINGS, msg.job_settings, 'settings', 0.02),
            (Job.PREP_INSTALL, msg.job_prep_install, 'prep_install', 0.80),
            (Job.INSTALL, msg.job_install, 'install', 0.02),
            (Job.POST_INSTALL, msg.job_post_install, 'post_install', 0.10),
            (Job.RUNTIME_CONF, msg.job_runtime_conf, 'runtime_conf', 0.02),
            (Job.RUNTIME, msg.job_runtime, 'runtime', 0.02),
            (Job.FINISHED, msg.job_finished, 'finish', 0.02),
        )
        self._guest = None
        self._state_num = Job.SETTINGS
        self._options = None
        self._xml_conf = self._xen_conf = None
        self._cancel = False
        self._vncport = None

        signal.signal(signal.SIGUSR1, self._handle_cancel_signal)

        self.background = background
        if self.background:
            (self.jobfd, self.jobfn) = tempfile.mkstemp(prefix=paths.jobdir)
            self.jobid = self.jobfn.split('/')[-1]
            print self.jobid
            sys.stdout.flush()
            def init():
                self.job_exit = None
                (_, self.logfn) = tempfile.mkstemp(prefix=paths.logdir+'/vm-install.', suffix='.log')
                self.pid = os.getpid()
                self._update_job_file()
            if not util.daemonize(cb=init):
                os._exit(0)
            init_log(debug=debug, filename=self.logfn)
        else:
            init_log(debug=debug)
            self.pid = os.getpid()
            self.jobfd = self.jobfn = self.logfn = self.jobid = None

    def __del__(self):
        log.debug("Job.__del__")
        #signal.signal(signal.SIGUSR1, signal.SIG_DFL)
        self._cleanup_conf()
        self.pid = None
        if self._state_num not in (Job.FINISHED, Job.FAILED, Job.CANCELED):
            log.info("Job exiting while in state %s; assuming canceled" % str(self._state_num))
            self.job_exit = JobError(JobError.E_CANCELED).exit_code()
            self._state_num = Job.CANCELED
        self._update_job_file()

    def _update_job_file(self):
        if not self.background:
            return
        state_num = self._state_num
        s = ''
        if self._vncport is not None:
            s += 'vnc ::%d\n' % self._vncport
        if self._options and self._options.uuid:
            s += 'uuid %s\n' % self._options.uuid
        if self._options and self._options.name:
            s += 'name %s\n' % self._options.name
        if self.job_exit is not None:
            s += 'exit %d\n' % self.job_exit
        if self.pid:
            s += 'pid %d\n' % self.pid
        s += 'id %s\nlog %s\nstate %d\ninotify %s\n' % (self.jobid, self.logfn, state_num, self.jobfn)
        if self.jobfd is None:
            self.jobfd = os.open(self.jobfn, os.O_WRONLY | os.O_TRUNC)
        os.write(self.jobfd, s)
        os.close(self.jobfd)
        self.jobfd = None

    def _cleanup_conf(self):
        if self._state_num == Job.FINISHED:
            return
        if self._xml_conf:
            log.info("Removing configuration file %s" % self._xml_conf)
            try:
                os.unlink(self._xml_conf)
            finally:
                self._xml_conf = None
        if self._xen_conf:
            log.info("Removing configuration file %s" % self._xen_conf)
            try:
                os.unlink(self._xen_conf)
            finally:
                self._xen_conf = None

    def show_console(self):
        if (not self._options.noautoconsole) and self._ConsoleCB is not None:
            self._ConsoleCB(self)
        else:
            wait_on_me = self._show_console()
            if type(wait_on_me) is str:
                console.wait_for_uuid_to_exit(wait_on_me)
            elif type(wait_on_me) is int:
                (pid, status) = os.waitpid(wait_on_me, 0)

    def _show_console(self):
        """Displays a console to the user, and if appropriate, waits for the
        console to close before continuing."""
        detached = (self.state[0] != Job.INSTALL)
        if self._options.noautoconsole:
            log.info('Not starting a console due to --no-autoconsole')
            domid = console.get_domid_from_uuid(self._options.uuid)
            self._vncport = console.get_vnc_port_from_domid(domid, timeout=10.0)
            log.debug('VNC port is %s' % self._vncport)
            if self._vncport is not None:
                self._update_job_file()
            if not detached:
                console.wait_for_domid_to_exit(domid, self._options.uuid)
        else:
            if detached:
                log.info('Starting a console, but not waiting for it to finish')
            else:
                log.info('Starting a console, and will wait for it to finish')
            child = None
            try_text = False
            if self._options.graphics == 'none':
                try_text = True
            else:
                if self._options.graphics_viewer == 'sdl':
                    # xm automatically created a viewer, so just wait.
                    log.debug('Viewer is sdl')
                elif self._options.graphics_viewer == 'vnc':
                    log.debug('Viewer is vnc')
                    try:
                        child = console.vnc_console(self.guest, detach=detached)
                    except NotImplementedError:
                        try_text = True
                    except RuntimeError, e:
                        self._ErrorCB(self, str(e))
            if try_text:
                log.debug('Viewer is text')
                try:
                    child = console.text_console(self.guest, detach=detached)
                except NotImplementedError:
                    self._ErrorCB(self, msg.display_failed)
            if child is None:
                if detached:
                    return None
                log.debug('Not detached but no child; wait for VM to exit')
                return self._options.uuid
            else:
                assert(not detached)
                log.debug('Not detached so waiting on PID')
                return child

    def run(self, GetSettingsCB, StateChangeCB=lambda job,state:None, ConsoleCB=None, \
            ErrorCB=lambda job,msg:None, ProgressCB=lambda job,f:None):
        """Starts the VM creation job.

        GetSettingsCB:  Called with job to get user's VM settings.
        StateChangeCB:  Called with (job, (state_N, of_M, description)) when state changes.
        ConsoleCB:  Called when console needs to be opened."""
        self._StateChangeCB = StateChangeCB
        self._GetSettingsCB = GetSettingsCB
        self._ConsoleCB = ConsoleCB
        self._ErrorCB = ErrorCB
        self._ProgressCB = ProgressCB
        self._cancel = False

        progress = util.filtered_progress(lambda f:self._ProgressCB(self, f))

        try:
            try:
                done = 0.0
                for state in self._states:
                    self._state_num = state[0]
                    self._update_job_file()
                    self._StateChangeCB(self, self.state)
                    func = eval('self._state_%s' % state[2])
                    func(lambda f:progress(done+f*state[3]))
                    done += state[3]
                    if self._cancel:
                        log.debug("Job's cancel flag was set")
                        raise JobError(JobError.E_CANCELED)
            except KeyboardInterrupt:
                raise JobError(JobError.E_CANCELED)
        except Error, e:
            log.info("vminstall.Error caught in job.run: %s" % str(e))
            if self._guest:
                self._guest.FailedInstall()
                self._guest = None
            if self.background:
                self.job_exit = e.exit_code()
                self._state_num = Job.FAILED
                self._update_job_file()
            raise
        except Exception, e:
            log.info("Exception caught in job.run: %s" % str(e))
            if self._guest:
                self._guest.FailedInstall()
                self._guest = None
            if self.background:
                self.job_exit = 1
                self._state_num = Job.FAILED
                self._update_job_file()
            raise

    def _handle_cancel_signal(self, signum, frame):
        log.debug("handle_cancel_signal")
        self.do_cancel(forceful=True)

    def signal_cancel(self):
        """Attempts to cancel the job.  Not safe to call in threaded environment."""
        os.kill(self.pid, signal.SIGUSR1)

    def do_cancel(self, forceful=False):
        """Attempts to cancel the job.  Safe to call from another thread."""
        if not forceful and self._state_num == Job.FINISHED:
            return
        self._cancel = True
        # Kill VM to break out of main job loop.
        # Beware: self._guest.uuid may not exist, or may disappear.
        #         Hold reference before testing.
        guest = self._guest
        if guest:
            uuid = guest.uuid
            if uuid:
                log.debug("Destroying VM %s" % uuid)
                xend.connection.destroy(uuid)

    def get_guest(self):
        return self._guest
    guest = property(get_guest)

    def get_state(self):
        return (self._state_num, len(self._states), self._states[self._state_num][1])
    state = property(get_state)

    def _state_settings(self, progress):
        if not caps.is_dom0():
            raise XenError(XenError.E_NO_XEN)
        if os.geteuid() != 0:
            raise XenError(XenError.E_ROOT)

        self._options = self._GetSettingsCB(self)
        progress(0.5)

        if self._options.full_virt:
            self._guest = FullVirtGuest(self._options)
        else:
            self._guest = ParaVirtGuest(self._options)

        for nic in self._options.nics:
            nic.setup()

        progress(1.0)

    def _state_prep_install(self, progress):
        if not self._options.install:
            return

        self.guest.PreInstall(lambda f:progress(f*0.98))

        conf = self.guest.GetConfigXml()
        self._xml_conf = os.path.join(paths.configdir, self.guest.options.name + '.xml')
        f = open(self._xml_conf, "w+")
        f.write(conf)
        f.close()
        conf = self.guest.GetConfigXen()
        self._xen_conf = os.path.join(paths.configdir, self.guest.options.name)
        f = open(self._xen_conf, "w+")
        f.write(conf)
        f.close()
        progress(1.0)

    def _state_install(self, progress):
        if not self._options.install:
            return
        self.guest.Start()
        self.guest.manage_disks()
        if not self._cancel:
            progress(0.5)
            self.show_console()
        progress(0.9)
        self.guest.Stop()
        progress(1.0)

    def _state_post_install(self, progress):
        log.info("Post install check...")
        self.guest.PostInstallCheck()
        progress(1.0)

    def _state_runtime_conf(self, progress):
        log.info("Post install configuration...")
        self.guest.PostInstallConf()
        progress(0.9)
        log.info("Writing runtime configuration file...")
        conf = self.guest.GetConfigXen()
        cf = os.path.join(paths.configdir, self.guest.options.name)
        f = open(cf, 'w+')
        f.write(conf)
        f.close()
        progress(1.0)

    def _state_runtime(self, progress):
        if self._options.norestart:
            return
        self.guest.start_from_disk()
        progress(0.5)
        self.show_console()
        progress(1.0)

    def _state_finish(self, progress):
        self.job_exit = 0
        progress(1.0)


def _get_job_filename(job_id):
    if '/' in job_id:
        raise JobError(JobError.E_NOT_FOUND)
    return os.path.join(paths.jobdir, job_id)

def _get_status_text(job_id):
    try:
        fh = open(_get_job_filename(job_id))
        text = fh.read()
        fh.close()
        return text
    except:
        raise JobError(JobError.E_NOT_FOUND)

def _get_status(job_id):
    text = _get_status_text(job_id)
    status = {}
    attrs = ('exit', 'id', 'inotify', 'log', 'pid', 'state', 'uuid', 'name', 'vnc')
    for attr in attrs:
        status[attr] = None
    for line in text.splitlines():
        words = line.split()
        if not words[0] in attrs:
            assert(False)
            continue
        status[words[0]] = words[1]
    for attr in ('exit', 'pid', 'state'):
        if status[attr] is not None:
            status[attr] = int(status[attr])
    # Possible (although unlikely) that the state file is truncated.
    # Ensure at least 'id' is set, to avoid needless exceptions.
    status['id'] = job_id
    return status

def try_unlink(file):
    try:
        os.unlink(file)
    except:
        pass

def _delete(status):
    if status['name'] is not None:
        try_unlink(os.path.join(paths.configdir, status['name']))
    if status['uuid'] is not None:
        xend.connection.delete(status['uuid'])
    try_unlink(_get_job_filename(status['id']))
    try_unlink(status['log'])

def list():
    return filter(lambda fn: len(fn) == 6, os.listdir(paths.jobdir))

def cancel(job_id):
    status = _get_status(job_id)
    if status['pid'] is not None:
        try:
            os.kill(status['pid'], signal.SIGUSR1)
        except:
            pass
    if status['uuid'] is not None:
        if xend.connection.domain_running(status['uuid']):
            xend.connection.destroy(status['uuid'])

def status(job_id):
    return _get_status_text(job_id)

def detach(job_id):
    status = _get_status(job_id)
    if status['state'] != Job.FINISHED:
        raise JobError(JobError.E_JOB_STATE)
    try_unlink(_get_job_filename(job_id))
    try_unlink(status['log'])

def delete(job_id):
    status = _get_status(job_id)
    if status['state'] not in (Job.FINISHED, Job.FAILED, Job.CANCELED):
        raise JobError(JobError.E_JOB_STATE)
    _delete(status)
